"""argdeco.arguments -- manage arguments


"""

import logging, inspect
logger = logging.getLogger('argparse.arguments')

from argparse import Action


class ArgAction(Action):
    '''Internal class to handle argument actions

    There are two ways

    '''

    def set_arg_func(self, arg_func):
        self.arg_func = arg_func

    def __call__(self, parser, namespace, values, option_string=None):
        if self.arg_func.__code__.co_argcount == 1:
            setattr(namespace, self.dest, self.arg_func(values))
        else:
            self.arg_func(self, parser, namespace, values, option_string)


class arg(object):
    """Represent arguments passed with add_argument() to an argparser

    See https://docs.python.org/3/library/argparse.html#argparse.ArgumentParser.add_argument
    """
    def __init__(self, *args, **opts):
        if 'config' in opts:
            self.config_name = opts.pop('config')
        if 'config_name' in opts:
            self.config_name = opts.pop('config_name')

        self.args = args
        self.opts = opts

    def apply(self, parser, command, context=''):
        if hasattr(self, 'config_name'):
            config_name = self.config_name
        else:
            config_name = '.'.join([context, self.dest])

        logger.info("do register_config_map: context=%s, self.dest=%s, config_name=%s", context, self.dest, config_name)
        command.register_config_map(context, self.dest, config_name)

        logger.info("apply: %s", self)
        parser.add_argument(*self.args, **self.opts)

    def __getattr__(self, name):
        if name == 'dest':
            dest = self.opts.get('dest')

            if not dest:
                for a in self.args:
                    if a.startswith('--'):
                        dest = a[2:].replace('-', '_')
                        break
                if not dest:
                    for a in self.args:
                        if a.startswith('-'):
                            dest = a[1:].replace('-', '_')
                            break

                if not dest:
                    dest = self.args[0]

            self.dest = dest
            return self.dest

        # if name == 'config_name':
        #     import rpdb2 ; rpdb2.start_embedded_debugger('foo')
        #
        #     path = [self.dest]
        #     cmd = self.command
        #     while cmd:
        #         if cmd.name:
        #             path.append(cmd.name)
        #         cmd = cmd.parent
        #     self.config_name = '.'.join(reversed(path))
        #     return self.config_name

        raise AttributeError(name)

    #     for a in self.args:
    #         if a.startswith('')
    #
    #     cmd =
    #     self.command

    def __repr__(self):
        return "arg(%s, %s)" % (self.args, self.opts)

    def __call__(self, *args, **kwargs):
        # factory for other arg
        if not (len(args) == 1 and inspect.isfunction(args[0])):
            _args = self.args
            if len(args):
                _args = args
            _opts = self.opts.copy()
            _opts.update(**kwargs)
            return arg(*_args, **_opts)

        func = args[0]
        self.func = func

        def arg_action_factory(*args, **kwargs):
            a = ArgAction(*args, **kwargs)
            a.set_arg_func(func)
            return a

        self.opts['action'] = arg_action_factory

        return self


class opt(arg):
    """Option action="store_true" """

    def apply(self, parser, command, context=''):
        logger.info("apply: %s", self)
        self.opts['action'] = 'store_true'
        self.opts['default'] = False
        arg.apply(self, parser, command, context)


class group(arg):
    """Argument group

    This class is a wrapper for
    :py:meth:`argparse.ArgumentParser.add_argument_group`.

    Usage::

        @main(
            group(
                arg('--first'),
                arg('--second'),
                title="group title",
                description='''
                   Here some group description
                '''
            )
        )
        def _main(first, second):
            pass
    """

    def apply(self, parser, command, context='', method='add_argument_group'):
        more_args = self.opts.pop('args', [])
        group = getattr(parser, method)(**self.opts)

        for a in self.args:
            a.apply(group, command)
        for a in more_args:
            a.apply(group, command)


class mutually_exclusive(group):
    """Mutually exclusive argument group

    Usage::

        @main(
            mutually_exclusive(
                arg('--first'),
                arg('--second'),
                title="group title",
                description='''
                   Here some group description
                '''
            )
        )
        def _main(first, second):
            pass

    """

    def apply(self, parser, command, context=''):
        group.apply(self, parser, command, context, 'add_mutually_exclusive_group')



"""argdeco -- use argparse with decorators

This module is main user interface.

Quickstart
----------

If you want to create a simple program::

    from argdeco import main, arg, opt

    @main(
        arg('--first', help="first argument"),
        opt('--flag-1', help="toggle 1st thing"),
        opt('--flag-2', help="toggle 2nd thing"),
    )
    def my_main(first, flag_1, flag_2):
        # do something here
        pass

    if __name__ == "__main__":
        main()

If you want to create a program with subcommands::

    from argdeco import main, command, arg, opt

    @command('cmd1')
    def cmd1(global_arg):
        print("this is the first command")

    @command("cmd2", arg("--foo-bar"))
    def cmd2(global_arg, foo_bar):
        print("this is the second command with arg: %s" % foo_bar)

    if __name__ == '__main__':
        main(arg('--global-arg', help="a global arg applied to all commands"))

Compiling arguments
-------------------

.. highlight:: py

If you have many arguments it may get cumbersome to list all the arguments in
the decorator and the function::

    @command('cmd1',
        arg('--first', '-f'),
        arg('--second', '-s'),
        arg('--third', '-t'),
        arg('--fourth', '-F'),
    )
    def cmd1(first, second, third, fourth):
        pass

Think of having many of such functions maybe even repeating some arguments.
Then it becomes handy to use compiled arguments::

    from argdeco import main, command

    @command('cmd',
        arg('--first', '-f'),
        arg('--second', '-s'),
        arg('--third', '-t'),
        arg('--fourth', '-F'),
    )
    def cmd1(opts):
        if opts['first'] == '1':
            ...

    if __name__ == '__main__':
        main(compile=True)

``compile`` can have following values:

===========  ===========  =====================================================
  Value        Alias       Description
===========  ===========  =====================================================
  None        'kwargs'    Passed to handler as keyword arguments
  True        'dict'      Args passed to handler as single dictionary
  'args'                  Pass args namespace as returned from :py:meth:`argparse.ArgumentParser.parse_args`
  function                You can also pass a function, which is explained in `Compile functions`_
===========  ===========  =====================================================

Compile to args::

    from argdeco import main, command

    @command('cmd',
        arg('--first', '-f'),
        arg('--second', '-s'),
        arg('--third', '-t'),
        arg('--fourth', '-F'),
    )
    def cmd1(args):
        if args.first == '1':
            ...

    if __name__ == '__main__':
        main(compile='args')


Compile functions
'''''''''''''''''

If you need even more control of your arguments, you can pass custom compile
functions, which gets args namespace and opts keyword arguments as parameter
and is expected to return:

================  =================================================================
   type           description
================  =================================================================
   dict           This will be passed as keyword arguments to handler function
   tuple/list     A tuple with two values, a list (or tuple) and a dictionary,
                  which are passed as args and kwargs to handler.
   list/tuple     If the tuple or list does not match requirements in above,
                  it is assumed, that no kwargs shall be passe and this is the
                  args list for positional parameters.
================  =================================================================

You can use such a function work preprocessing some args and manipulate the
parameters passed to the handlers.




Compiler factory
''''''''''''''''

Compile functions are usually not directly connected with command decorator and
usually do not know about it (unless you share it globally).  If you need to
access data from command decorator instance or need for other reasons more
control of argument setup, you can use a compiler factory.

A compiler factory is initialized with :py:class:`CommandDecorator` instance.

It must return a function, which will get args as returned from
:py:meth:`argparse.ArgumentParser.parse_args` and keyword arguments.

Here you see the most simplest one::

    def my_factory(command):
        def my_compiler(args, **opts):
            return opts

        return my_compiler


Validating and transforming arguments
-------------------------------------

With :py:class:`argparse.Action` :py:mod:`argparse` module provides a method
to provide custom argument handlers.  :py:mod:`argdeco` provides some eases
for this as well::

    from argdeco import arg, command, main
    import dateutil.parser

    @arg('--date', '-d')
    def arg_date(value):
        return dateutil.parser.parse(value)

    @main(arg_date)
    def handle_date(date)
        print(date.isoformat())

    main()


There is also a complex (more powerful) way, which is ::

    import dateutil.parser

    from argdeco import arg, command, main

    @arg("-d", "--date", help="pass some date")
    def arg_date(self, parser, namespace, values, option_string=None):
         # here we can do some validations
         print "self: %s" % self
         setattr(namespace, self.dest, dateutil.parser.parse(values))

    @command("check_date", arg_date)
    def check_date(date):
        print(date)

    main()


Working with subcommands
------------------------

You may want to implement a CLI like git has.  This is quite easy with
:py:mod:`argdeco`::

    from argdeco import main, command, arg, opt
    from textwrap import dedent

    # we will implement a sample `remote` command here

    # global arguments (for all commands)
    main(
        arg('--config-file', '-C', help="pass a config file"),
    )

    # create a new decorator for sub-command 'remote' actions
    remote_command = command.add_subcommands('remote',
        help="manage remote sites",  # description in global command list
        subcommands = dict(
            title = "remote commands",   # caption of subcommand list
            description = dedent('''\
                Here is some documentation about remote commands.

                There is a lot to say ...
            ''')
        )
    )

    @remote_command('add',
        arg('remote_name', help="name of remote site"),
        arg('url', help="url of remote site"),
        opt('--tags', help="get all tags when requesting remote site"),
    )
    def cmd_remote_add(config_file, remote_name, url, tags):
        ...

    @remote_command('rename',
        arg('old_name', help="old name of remote"),
        arg('new_name', help="new name of remote"),
    )
    def cmd_remote_rename(config_file, old_name, new_name):
        ...

If you run ``add_subcommands(..., subcommands={...})``, all the
keyword arguments of add_subcommands, except the ``subcommands`` one, will be passed
to :py:meth:`argparse.ArgumentParser.add_parser` and the subcommands dictionary
will be passed as keyword arguments to
:py:meth:`argparse.ArgumentParser.add_subparsers`.

"""
from .main import Main
from .command_decorator import CommandDecorator
from .arguments import arg, group, mutually_exclusive, opt
from .config import config_factory, ConfigDict

option = opt

main    = Main()
command = main.command

__version__ = '2.0.16'

#



from distutils.core import setup
from argdeco import __version__
from textwrap import dedent

setup(name='argdeco',
  version=__version__,
  #py_modules=['argdeco'],
  packages = ['argdeco'],
  author = 'Kay-Uwe (Kiwi) Lorenz',
  author_email = "kiwi@franka.dyndns.org",
  url = 'https://github.com/klorenz/python-argdeco',
  description = "specify command arguments in function decorator",
  long_description = dedent("""\
     argdeco
     =======

     Specify command line arguments using decorators::

         from argdeco import main, arg, opt

         @main(
            arg("--foo", help="some argument"),
            opt("--flag", '-f', help="toggle flag"),
         )
         def my_main_function(foo, flag):
             return 0  # success, will be exit code

         if __name__ == '__main__':
             main()

     argdeco is an argparse wrapper.

     * ``arg()`` is only a wrapper around ``argparse.ArgumentParser.add_argument()``.
     * ``opt()`` is a shorthand for ``arg(..., action=store_true, default=False)``

     ... you can do `much more`_

     .. [much more] https://python-argdeco.readthedocs.io

  """),
  keywords = "argument command argparse cli",
  #download_url
  license = "MIT",
)



"""Working with configurations

A common pattern making use of configuration files::

    from argdeco import main, command, arg, opt, config_factory
    form os.path import expanduser

    main.configure(compiler_factory=config_factory(
        config_file=arg('--config-file', '-C', help="configuration file", default=expanduser('~/.config/myconfig.yaml'))
    ))

    @command('ls', opt('--all'))
    def mycmd(cfg):
        if cfg['ls.all']:
            pass

    main()

If you want to have ``foo.bar`` expanded to ``{'foo': {'bar': ...}}``, use
following::

    from argdeco import main, command, arg, opt, config_factory, ConfigDict
    form os.path import expanduser

    main.configure(compiler_factory=config_factory(ConfigDict,
        config_file=arg('--config-file', '-C', help="configuration file", default=expanduser('~/.config/myconfig.yaml'))
    ))

    @command('ls', opt('--all'))
    def mycmd(cfg):
        if cfg['ls']['all']:
            pass

    main()


"""

import logging
log = logging.getLogger('argdeco.config')

class ConfigDict(dict):
    '''dictionary-like class

    This class implements a dictionary, which creates deep objects from keys
    like "foo.bar".  Example:

    >>> c = Config()
    >>> c['foo.bar'] = 'x'
    >>> c
    {'foo': {'bar': 'x'}}
    >>> c['foo.bar']
    'x'

    '''

    def __init__(self, E=None, **F):
        super(ConfigDict, self).__init__()
        self.update(E, **F)

    def __getitem__(self, name):
        key_parts = name.split('.')
        value = super(ConfigDict, self).__getitem__(key_parts[0])
        for k in key_parts[1:]:
            value = value[k]

        return value

    def __setitem__(self, name, value):
        key_parts = name.split('.')
        if len(key_parts) == 1:
            super(ConfigDict, self).__setitem__(key_parts[0], value)
        else:
            try:
                val = super(ConfigDict, self).__getitem__(key_parts[0])
            except KeyError:
                super(ConfigDict, self).__setitem__(key_parts[0], self.__class__())
                val = super(ConfigDict, self).__getitem__(key_parts[0])

            for k in key_parts[1:-1]:
                if k not in val:
                    val[k] = self.__class__()
                val = val[k]

            if isinstance(value, dict) and not isinstance(value, self.__class__):
                val[key_parts[-1]] = self.assimilate(value)
            else:
                val[key_parts[-1]] = value

    def assimilate(self, value):
        if not isinstance(value, dict):
            return value
        if isinstance(value, self.__class__):
            return value

        log.debug("assimilate %s", value)
        result = self.__class__()
        result.update(value)

        for a in dir(value):
            if not hasattr({}, a):
                log.debug("assimilate, %s => %s", a, getatt(value, a))
                setattr(result, a, getattr(value, a))

        return result

    def __contains__(self, name):
        try:
            self[name]
            return True
        except KeyError:
            return False


    def flatten(self, D):
        '''flatten a nested dictionary D to a flat dictionary

        nested keys are separated by '.'
        '''

        if not isinstance(D, dict):
            return D

        result = {}
        for k,v in D.items():
            if isinstance(v, dict):
                for _k,_v in self.flatten(v).items():
                    result['.'.join([k,_k])] = _v
            else:
                result[k] = v
        return result


    def update(self, E=None, **F):
        '''flatten nested dictionaries to update pathwise

        >>> Config({'foo': {'bar': 'glork'}}).update({'foo': {'blub': 'bla'}})
        {'foo': {'bar': 'glork', 'blub': 'bla'}

        In contrast to:

        >>> {'foo': {'bar': 'glork'}}.update({'foo': {'blub': 'bla'}})
        {'foo: {'blub': 'bla'}'}

        '''
        def _update(D):
            for k,v in D.items():
                if super(ConfigDict, self).__contains__(k):
                    if isinstance(self[k], ConfigDict):
                        self[k].update(v)
                    else:
                        self[k] = self.assimilate(v)
                else:
                    self[k] = self.assimilate(v)

        if E is not None:
            if not hasattr(E, 'keys'):
                E = self.assimilate(dict(E))
            _update(E)

        _update(F)

        return self



def config_factory(ConfigClass=dict, prefix=None,
    config_file=None
    ):
    '''return a class, which implements the compiler_factory API

    :param ConfigClass:
        defaults to dict.  A simple factory (without parameter) for a
        dictionary-like object, which implements __setitem__() method.

        Additionally you can implement following methods:

        :``init_args``: A method to be called to initialize the config object
           by passing :py:class:`~argparse.Namespace` object resulting from
           :py:class:`~argparse.ArgumentParser.parseargs` method.

           You could load data from a configuration file here.

        :``compile_args``: A method, which can return the same like a
           ``compile`` function does.  If there is no such method, a tuple
           with a ConfigClass instance as single element is returned.

    :param prefix:
        Add this prefix to config_name.  (e.g. if prefix="foo" and you
        have config_name="x.y" final config_path results in "foo.x.y")

    :param config_file:
        An :py:class:`~argdeco.arguments.arg` to provide a config file.

        If you provide this argument, you can implement one of the following
        methods in your ``ConfigClass`` to load data from the configfile:

        :``load``: If you pass ``config_file`` argument, this method can be
           implemented to load configuration data from resulting stream.

           If config_file is '-', stdin stream is passed.

        :``load_from_file``: If you prefer to open the file yourself, you can
           do this, by implementing ``load_from_file`` instead which has the
           filename as its single argument.

        :``update``: method like :py:meth:`dict.update`.   If neither of
           ``load`` or ``load_from_file`` is present, but ``update`` is,
           it is assumed, that config_file is of type YAML (or JSON) and
           configuration is updated by calling ``update`` with the parsed data
           as parameter.

        If you implement neither of these, it is assumed, that configuration
        file is of type YAML (or plain JSON, as YAML is a superset of it).

        Data is loaded from file and will update configuration object using
        dict-like :py:meth:`dict.update` method.

    :type config_file: argdeco.arguments.arg

    :returns:
        ConfigFactory class, which implements compiler_factory API.
    '''
    config_factory = ConfigClass

    class ConfigFactory:
        def __init__(self, command):
            self.command = command
            if config_file:
                from .arguments import arg
                assert isinstance(config_file, arg), "config_file must be of type arg"
                try:
                    self.command.add_argument(config_file)
                except:
                    pass

        def __call__(self, args, **opts):
            cfg = ConfigClass()

            if hasattr(cfg, 'init_args'):
                cfg.init_args(args)

            if config_file is not None:
                if hasattr(args, config_file.dest):
                    fn = getattr(args, config_file.dest)
                    if fn is not None:
                        if hasattr(cfg, 'load'):
                            if config_file.dest == '-':
                                cfg.load(sys.stdin)
                            else:
                                with open(fn, 'r') as f:
                                    cfg.load(f)

                        elif hasattr(cfg, 'load_from_file'):
                            cfg.load_from_file(fn)

                        elif hasattr(cfg, 'update'):
                            # assume yaml file
                            import yaml
                            with open(fn, 'r') as f:
                                data = yaml.load(f)
                            cfg.update(data)

            for k,v in opts.items():
                config_name = self.command.get_config_name(args.action, k)

                if config_name is None: continue
                if prefix is not None:
                    config_name = '.'.join([prefix, config_name])
                cfg[config_name] = v

            if hasattr(cfg, 'compile_args'):
                return cfg.compile_args()
            else:
                return (cfg,)

    return ConfigFactory



from argparse import ArgumentParser
from textwrap import dedent
from .arguments import arg

import logging, sys, argparse

logger = logging.getLogger('argdeco.command_decorator')

try:
    basestring
except NameError:
    basestring = str

class Undefined:
    pass

class CommandDecorator:
    """
    Create a decorator to decorate functions with their arguments.
    """

    def __init__(self, *args, **kwargs):
        """Initialize CommandDecorator with global arguments

        >>> from argdeco import arg, CommandDecorator
        >>> command = CommandDecorator(
        ...
        ... )

        # Arguments

        description:
          Description in argument, also causes that command's function __doc__
          documentation will be also passed as description to add_argument()
          by default.

          If not specified, the calling modules __doc__ will be used as epilogue
          (and all other commands __doc__ will be passed as epilogue).

        formatter_class:
          Default formatter class is argparse.RawDescriptionHelpFormatter.  You
          can override this with passing another formatter_class.

        preprocessor:
          Instead of passing a compile paramter to execute later, you can also
          pass a preprocessor function here.

          Preprocessor function is getting args object as positional argument
          and expanded arguments as kwargs.

          It may return

          # a tuple of (args, kwargs) to be passed to command function
          # a dict, to be passed as kwargs to command function
          # a tuple or list (with not per accident having exact two items
            beeing a tuple and kwargs) passed as positional args to command
            function

        argparser:
          (internal) You may pass an existing argparser instance.

        commands:
          (internal) An object returned from argparser.add_subparsers() method.

        compiler_factory:
          Passing a compiler_factory is another way of providing a argument
          compiler.

          `compiler_factory` is expected to be a function getting one parameter
          -- the args as returned from `argparser.parseargs()``.  It is espected
          to return a dictionary like object.  For each given argument the
          name of the argument is computed and the corresponding configuration
          item is set.

          Example:

              >>> command = CommandDecorator(compiler_factory=lambda x: {})
              >>> @command('foo', arg('--bar'))
              >>> def cmd(cfg):
              ...     assert cfg['foo.bar'] == '1'
              >>> command.execute(['foo', '--bar', '1'])

        """
        self.formatter_class = kwargs.get('formatter_class', argparse.RawDescriptionHelpFormatter)

        if 'description' in kwargs:
            self.doc = 'description'
        else:
            self.doc = 'epilog'

        if 'description' not in kwargs and 'epilog' not in kwargs:
            kwargs[self.doc] = sys._getframe().f_back.f_globals.get('__doc__')

        if 'formatter_class' not in kwargs:
            kwargs['formatter_class'] = self.formatter_class

        if 'preprocessor' in kwargs:
            self.preprocessor = kwargs.pop('preprocessor')
        else:
            self.preprocessor = None

        # use epilog or description
        #moddoc = sys._getframe().f_back.f_globals.get('__doc__')
        #epilog =

        for k in ('commands', 'parent', 'name', 'compiler_factory'):
            if k in kwargs:
                setattr(self, k, kwargs.pop(k))
            else:
                setattr(self, k, None)

        self.config_map = {}
        self.compile = None

        if 'argparser' in kwargs:
            self.argparser = kwargs.pop('argparser')
        else:
            self.argparser = ArgumentParser(**kwargs)

        for a in args:
            if isinstance(a, arg):
                a.apply(self.argparser, self, self.get_name())

        self.children = []


    def has_action(self):
        if self.commands is not None:
            return len(self.commands._name_parser_map) > 0
        else:
            return False

    def add_parser(self, command, *args, **kwargs):
        if self.commands is None:
            self.commands = self.argparser.add_subparsers()

        return self.commands.add_parser(command, *args, **kwargs)

    def __getitem__(self, name):
        if self.commands is None:
            raise KeyError(name)

        cmd = self
        items = name.split('.')
        path, name = items[:-1], items[-1]

        for k in path:
            _map = dict((c.name, c) for c in cmd.children)
            if k not in _map:
                raise KeyError(name)
            cmd = _map[k]

        return cmd.commands._name_parser_map[name]

    def get_action(self, name):
        return self[name].get_default('action')

    def add_subcommands(self, command, *args, **kwargs):
        """add subcommands.

        If command already defined, pass args and kwargs to add_subparsers()
        method, else to add_parser() method.  This behaviour is for convenience,
        because I mostly use the sequence:

        >>> p = parser.add_parser('foo', help="some help")
        >>> subparser = p.add_subparsers()

        If you want to configure your sub_parsers, you can do it with:

        >>> command.add_subcommands('cmd',
                help = "cmd help"
                subcommands = dict(
                    title = "title"
                    description = "subcommands description"
                )
            )
        """
        subcommands = kwargs.pop('subcommands', None)
        try:
            cmd = self[command]
        except KeyError:
            if 'formatter_class' not in kwargs:
                kwargs['formatter_class'] = self.formatter_class

            cmd = self.add_parser(command, *args, **kwargs)
            args, kwargs = tuple(), dict()

        if subcommands is not None:
            kwargs = subcommands

        child = CommandDecorator(
            argparser = self.argparser,
            commands  = cmd.add_subparsers(*args, **kwargs),
            parent = self,
            name   = command,
            )
        self.children.append(child)

        return child

    def update(self, command=None, **kwargs):
        """update data, which is usually passed in ArgumentParser initialization

        e.g. command.update(prog="foo")
        """
        if command is None:
            argparser = self.argparser
        else:
            argparser = self[command]

        for k,v in kwargs.items():
            setattr(argparser, k, v)

    def add_argument(self, *args, **kwargs):
        logger.info("add_argument: %s, %s", args, kwargs)
        if len(args) == 1 and isinstance(args[0], arg):
            self.add_arguments(*args)
        else:
            self.argparser.add_argument(*args, **kwargs)

    # def register_config_map(self, dest):
    def register_config_map(self, context, dest, config_name):
        logger.info("register_config_map: context=%s, dest=%s, config_name=%s", context, dest, config_name)

        _root = self
        while _root.parent is not None:
            _root = _root.parent

        if context not in _root.config_map:
            _root.config_map[context] = {}

        _root.config_map[context][dest] = config_name

    #     map_name = self.get_name()
    #     logger.debug("map_name=%s", map_name)
    #     if hasattr(a, 'config_name'):
    #         config_name = a.config_name
    #     else:
    #         config_name = '.'.join([map_name, a.dest])
    #
    #     if map_name not in self.config_map:
    #         self.config_map[map_name] = {}


    def add_arguments(self, *args, **kwargs):
        context = kwargs.get('context')
        argparser = kwargs.get('argparser')
        if context is None:
            context = self.get_name()

        if argparser is None:
            argparser = self.argparser

        for a in args:
            if isinstance(a, basestring):
                a = arg(a)
            elif isinstance(a, dict):
                a = arg(**a)
            elif isinstance(a, (tuple,list)):
                a = arg(*a)
            elif not isinstance(a, arg):
                raise ValueError("cannot convert %s into arg type" % a)

            a.apply(argparser, self, context=context)


    def get_config_name(self, action, name=None):
        '''get the name for configuration

        This returns a name respecting commands and subcommands.  So if you
        have a command name "index" with subcommand "ls", which has option
        "--all", you will pass the action for subcommand "ls"  and the options's
        dest name ("all" in this case), then this function will return
        "index.ls.all" as configuration name for this option.
        '''

        _name = None

        if name is None:
            if '.' in action:
                action, name = action.rsplit('.', 1)
            else:
                _name = ''
                name = action

        if _name is None:
            if isinstance(action, basestring):
                action = self.get_action(action)
            _name = action.argdeco_name

        logger.debug("_name=%s", _name)

        config_name = Undefined
        while True:
            logger.debug("check _name=%s, name=%s", repr(_name), repr(name))

            if _name in self.config_map:
                if name in self.config_map[_name]:
                    config_name = self.config_map[_name][name]
                    if config_name is not None:
                        if config_name.startswith('.'):
                            config_name = _name + config_name
                    break

            if _name == '':
                break

            if '.' not in _name:
                _name = ''
                continue

            _name = _name.rsplit('.', 1)[0]

        assert config_name is not Undefined, "could not determine config name for %s" % name
        return config_name


    def add_command(self, command, *args, **kwargs):
        """add a command.

        This is basically a wrapper for add_parser()
        """
        cmd = self.add_parser(command, *args, **kwargs)


    def get_name(self, name=None):
        if name is not None:
            path = [name]
        else:
            path = []

        cmd = self
        while cmd:
            if cmd.name:
                path.append(cmd.name)
            cmd = cmd.parent

        return '.'.join(reversed(path))

    def __call__(self, *args, **opts):
        def factory(func):
            _args = args
            if func.__doc__ is not None:
                _doc = dedent(func.__doc__)

                if 'help' not in opts:
                    try:
                        help, desc = _doc.split("\n\n", 1)
                        help = help.strip()

                    except:
                        help = _doc.strip()
                        desc = ''
                else:
                    desc = _doc
            else:
                help = None
                desc = None

            kwargs = {
                'help': help,
                'formatter_class': self.formatter_class,
                self.doc: desc,
            }

            kwargs.update(opts)

            if len(_args) and isinstance(_args[0], basestring):
                name = _args[0]
                _args = _args[1:]

                command = self.add_parser(name, **kwargs)
            else:
                name = None
                command = self.argparser

            func.argdeco_name = context = self.get_name(name)
            self.config_map[func.argdeco_name] = {}

            self.add_arguments(*_args, argparser=command, context=context)

            command.set_defaults(action=func)

            return func

        return factory

    def compile_args(self, argv=None, compile=None, preprocessor=None, compiler_factory=None):
        if argv is None:
            argv = sys.argv[1:]

        # initialize arg processors
        if not preprocessor:
            preprocessor = self.preprocessor

        if not compile:
            compile = self.compile

        if not compiler_factory:
            compiler_factory = self.compiler_factory

        assert not (compiler_factory and compile), \
            "you can either define a compiler factory or a compile function"

        args = self.argparser.parse_args(argv)

        if preprocessor:
            result = preprocessor(args)

            # if preprocessor returns a value, this overrules everything
            # (e.g. print error message and return exit value)
            if result is not None:
                return result

        opts = vars(args).copy()
        del opts['action']

        if compiler_factory:
            compile = compiler_factory(self)

        if compile is None or compile == 'kwargs':
            (_args, _kwargs) = tuple(), opts
        elif compile is True or compile == 'dict':
            (_args, _kwargs) = (opts,), dict()
        elif compile == 'args':
            (_args, _kwargs) = (args,), dict()
        else:
            compiled = compile(args, **opts)

            if isinstance(compiled, dict):
                (_args, _kwargs) = tuple(), compiled
            elif isinstance(compiled, (tuple,list)) and len(compiled) == 2 and isinstance(compiled[1], dict) and isinstance(compiled[0], (tuple,list)):
                (_args, _kwargs) = compiled
            elif isinstance(compiled, (tuple,list)):
                (_args, _kwargs) = compiled, dict()
            else:
                raise "Unkown compilation: %s" % compiled

        logger.debug("compiled: %s", (_args, _kwargs))
        return (args.action, _args, _kwargs)


    def execute(self, argv=None, compile=None, preprocessor=None, compiler_factory=None):
        """Parse arguments and execute decorated function

        argv: list of arguments
        compile:
            - None, pass args as keyword args to function
            - True, pass args as single dictionary
            - function, get args from parse_args() and return a pair of
              tuple and dict to be passed as args and kwargs to function
        """

        action, args, kwargs = self.compile_args(argv=argv, compile=compile, preprocessor=preprocessor, compiler_factory=compiler_factory)
        return action(*args, **kwargs)


def factory(**kwargs):
    frame = sys._getframe()

    while 'argdeco' in frame.f_globals.get('__name__', ''):
        frame = frame.f_back

    doc = frame.f_globals.get('__doc__')
    if doc is not None and 'epilog' not in kwargs and 'description' not in kwargs:
        kwargs['epilog'] = doc
    return CommandDecorator(**kwargs)

#command_inst = None
